"""Package containig various model files for fitting.

This package contains the various MRI models currently implemented for the
toolbox. In additon, the "GeneralModel" can be run with a simple text file
to devine new model. An exemplary textfile for simple models can be
generated by running the "genDefaultModelfile" function of the "GeneralModel".
"""
# __all__ = ["BiExpDecay",
#            "DiffdirLL",
#            "GeneralModel",
#            "ImageReco",
#            "IRLL",
#            "template",
#            "VFA"]
from .BiExpDecay import Model as BiExpDecay
from .DiffdirLL import Model as DiffdirLL
from .GeneralModel import Model as GeneralModel
from .ImageReco import Model as ImageReco
from .IRLL import Model as IRLL
from .VFA import Model as VFA



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module holding the classes for IRGN Optimization without streaming."""
from __future__ import division
import time
import numpy as np

from pkg_resources import resource_filename
import pyopencl.array as clarray
import h5py

import pyqmri.operator as operator
import pyqmri.solver as optimizer
from pyqmri._helper_fun import CLProgram as Program
from pyqmri._helper_fun import _utils as utils


class IRGNOptimizer:
    """Main IRGN Optimization class.

    This Class performs IRGN Optimization either with TGV or TV regularization.

    Parameters
    ----------
      par : dict
        A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      model : pyqmri.model
        Which model should be used for fitting.
        Expects a pyqmri.model instance.
      trafo : int, 1
        Select radial (1, default) or cartesian (0) sampling for the fft.
      imagespace : bool, false
        Perform the fitting from k-space data (false, default) or from the
        image series (true)
      SMS : int, 0
        Select if simultaneous multi-slice acquisition was used (1) or
        standard slice-by-slice acquisition was done (0, default).
      reg_type : str, "TGV"
        Select between "TGV" (default) or "TV" regularization.
      config : str, ''
        Name of config file. If empty, default config file will be generated.
      streamed : bool, false
        Select between standard reconstruction (false)
        or streamed reconstruction (true) for large volumetric data which
        does not fit on the GPU memory at once.
      DTYPE : numpy.dtype, numpy.complex64
        Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
        Real working precission.

    Attributes
    ----------
      par : dict
        A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      gn_res : list of floats
        The residual values for of each Gauss-Newton step. Each iteration
        appends its value to the list.
      irgn_par : dict
        The parameters read from the config file to guide the IRGN
        optimization process
    """

    def __init__(self, par, model, trafo=1, imagespace=False, SMS=0,
                 reg_type='TGV', config='', streamed=False,
                 DTYPE=np.complex64, DTYPE_real=np.float32):
        self.par = par
        self.gn_res = []
        self.irgn_par = utils.read_config(config, reg_type)
        utils.save_config(self.irgn_par, par["outdir"], reg_type)
        num_dev = len(par["num_dev"])
        self._fval_old = 0
        self._fval = 0
        self._fval_init = 0
        self._ctx = par["ctx"]
        self._queue = par["queue"]
        self._model = model
        self._reg_type = reg_type
        self._prg = []

        self._DTYPE = DTYPE
        self._DTYPE_real = DTYPE_real

        self._streamed = streamed
        self._imagespace = imagespace
        self._SMS = SMS
        if streamed and par["NSlice"]/(num_dev*par["par_slices"]) < 2:
            raise ValueError(
                "Number of Slices devided by parallel "
                "computed slices and devices needs to be larger two.\n"
                "Current values are %i total Slices, %i parallel slices and "
                "%i compute devices."
                % (par["NSlice"], par["par_slices"], num_dev))
        if streamed and par["NSlice"] % par["par_slices"]:
            raise ValueError(
                "Number of Slices devided by parallel "
                "computed slices needs to be an integer.\n"
                "Current values are %i total Slices with %i parallel slices."
                % (par["NSlice"], par["par_slices"]))
        if DTYPE == np.complex128:
            if streamed:
                kernname = 'kernels/OpenCL_Kernels_double_streamed.c'
            else:
                kernname = 'kernels/OpenCL_Kernels_double.c'
            for j in range(num_dev):
                self._prg.append(Program(
                    self._ctx[j],
                    open(
                        resource_filename(
                            'pyqmri', kernname)
                        ).read()))
        else:
            if streamed:
                kernname = 'kernels/OpenCL_Kernels_streamed.c'
            else:
                kernname = 'kernels/OpenCL_Kernels.c'
            for j in range(num_dev):
                self._prg.append(Program(
                    self._ctx[j],
                    open(
                        resource_filename(
                            'pyqmri', kernname)).read()))

        if imagespace:
            self._coils = []
            self.sliceaxis = 1
            if self._streamed:
                self._data_trans_axes = (1, 0, 2, 3)
                self._grad_trans_axes = (2, 0, 1, 3, 4)
        else:
            if SMS:
                self._data_shape = (par["NScan"], par["NC"],
                                    par["packs"]*par["numofpacks"],
                                    par["Nproj"], par["N"])
            else:
                if par["is3D"] and trafo:
                    self._data_shape = (par["NScan"], par["NC"],
                                        1, par["Nproj"], par["N"])
                else:
                    self._data_shape = (par["NScan"], par["NC"],
                                        par["NSlice"], par["Nproj"], par["N"])
            if self._streamed:
                self._data_trans_axes = (2, 0, 1, 3, 4)
                self._grad_trans_axes = (2, 0, 1, 3, 4)
                self._coils = np.require(
                    np.swapaxes(par["C"], 0, 1), requirements='C',
                    dtype=DTYPE)

                if SMS:
                    self._data_shape = (par["packs"]*par["numofpacks"],
                                        par["NScan"],
                                        par["NC"], par["dimY"], par["dimX"])
                    self._data_shape_T = (par["NScan"], par["NC"],
                                          par["packs"]*par["numofpacks"],
                                          par["dimY"], par["dimX"])
                    self._expdim_dat = 1
                    self._expdim_C = 0
                else:
                    self._data_shape = (par["NSlice"], par["NScan"], par["NC"],
                                        par["Nproj"], par["N"])
                    self._data_shape_T = self._data_shape
                    self._expdim_dat = 2
                    self._expdim_C = 1
            else:
                self._coils = clarray.to_device(self._queue[0],
                                                self.par["C"])

        self._MRI_operator, self._FT = operator.Operator.MRIOperatorFactory(
            par,
            self._prg,
            DTYPE,
            DTYPE_real,
            trafo,
            imagespace,
            SMS,
            streamed
            )

        self._grad_op, self._symgrad_op, self._v = self._setupLinearOps(
            DTYPE,
            DTYPE_real)

        if not reg_type == "H1":
            self._pdop = optimizer.PDBaseSolver.factory(
                self._prg,
                self._queue,
                self.par,
                self.irgn_par,
                self._fval_init,
                self._coils,
                linops=(self._MRI_operator, self._grad_op, self._symgrad_op),
                model=model,
                reg_type=self._reg_type,
                SMS=self._SMS,
                streamed=self._streamed,
                imagespace=self._imagespace,
                DTYPE=DTYPE,
                DTYPE_real=DTYPE_real
                )
        else:
            self._pdop = optimizer.CGSolver_H1(
                self._prg,
                self._queue,
                self.par,
                self.irgn_par,
                self._coils,
                linops=(self._MRI_operator, self._grad_op)
                )

        self._gamma = None
        self._delta = None
        self._omega = None
        self._step_val = None
        self._modelgrad = None

    def _setupLinearOps(self, DTYPE, DTYPE_real):
        grad_op = operator.Operator.GradientOperatorFactory(
            self.par,
            self._prg,
            DTYPE,
            DTYPE_real,
            self._streamed)
        symgrad_op = None
        v = None
        if self._reg_type == 'TGV':
            symgrad_op = operator.Operator.SymGradientOperatorFactory(
                self.par,
                self._prg,
                DTYPE,
                DTYPE_real,
                self._streamed)
            v = np.zeros(
                ([self.par["unknowns_TGV"], self.par["NSlice"],
                  self.par["dimY"], self.par["dimX"], 4]),
                dtype=DTYPE)
            if self._streamed:
                v = np.require(np.swapaxes(v, 0, 1), requirements='C')
        return grad_op, symgrad_op, v

    def execute(self, data):
        """Start the IRGN optimization.

        This method performs iterative regularized Gauss-Newton optimization
        and calls the inner loop after precomputing the current linearization
        point. Results of the fitting process are saved after each
        linearization step to the output folder.

        Parameters
        ----------
          data : numpy.array
            the data to perform optimization/fitting on.
        """
        # self.irgn_par["lambd"] *= (
        #                             (self.par["SNR_est"]))
        self._gamma = self.irgn_par["gamma"]
        self._delta = self.irgn_par["delta"]
        self._omega = self.irgn_par["omega"]

        iters = self.irgn_par["start_iters"]
        result = np.copy(self._model.guess)

        if "inc" in self.irgn_par.keys():
            inc = int(self.irgn_par["inc"])
        else:
            inc = 2

        if self._streamed:
            data = np.require(
                np.transpose(data, self._data_trans_axes),
                requirements='C')

        self._step_val = np.nan_to_num(self._model.execute_forward(result))

        if self._streamed:
            if self._SMS is False:
                self._step_val = np.require(
                    np.swapaxes(self._step_val, 0, 1), requirements='C')

        for ign in range(self.irgn_par["max_gn_it"]):
            start = time.time()
            self._modelgrad = np.nan_to_num(
                self._model.execute_gradient(result))

            self._balanceModelGradients(result, ign)
            self._updateIRGNRegPar(ign)
            # self._pdop._grad_op.updateRatio(
            #     self.irgn_par["gamma"])
            # if self._reg_type == 'TGV':
            #     self._pdop._symgrad_op.updateRatio(
            #         2*self.irgn_par["gamma"])

            self._step_val = np.nan_to_num(self._model.execute_forward(result))

            if self._streamed:
                if self._SMS is False:
                    self._step_val = np.require(
                        np.swapaxes(self._step_val, 0, 1), requirements='C')
                self._modelgrad = np.require(
                    np.transpose(self._modelgrad, self._grad_trans_axes),
                    requirements='C')
                self._pdop.model = self._model
                self._pdop.modelgrad = self._modelgrad
                self._pdop.jacobi = np.sum(
                    np.abs(self._modelgrad)**2, 2).astype(self._DTYPE_real)
                self._pdop.jacobi[self._pdop.jacobi == 0] = 1e-8
            else:
                _jacobi = np.sum(
                    np.abs(
                        self._modelgrad)**2, 1).astype(self._DTYPE_real)
                _jacobi[_jacobi == 0] = 1e-8
                self._modelgrad = clarray.to_device(
                    self._queue[0],
                    self._modelgrad)
                self._pdop.model = self._model
                self._pdop.modelgrad = self._modelgrad
                self._pdop.jacobi = clarray.to_device(
                    self._queue[0],
                    _jacobi)

            self._pdop.updateRegPar(self.irgn_par)

            result = self._irgnSolve3D(result, iters, data, ign)

            iters = np.fmin(iters * inc, self.irgn_par["max_iters"])

            end = time.time() - start
            self.gn_res.append(self._fval)
            print("-" * 75)
            print("GN-Iter: %d  Elapsed time: %f seconds" % (ign, end))
            print("-" * 75)
            self._fval_old = self._fval
            self._saveToFile(ign, self._model.rescale(result)["data"])
        self._calcResidual(result, data, ign+1)

    def _updateIRGNRegPar(self, ign):
        try:
            self.irgn_par["delta"] = np.minimum(
                self._delta
                * self.irgn_par["delta_inc"]**ign,
                self.irgn_par["delta_max"])*self.irgn_par["lambd"]
        except OverflowError:
            self.irgn_par["delta"] = np.minimum(
                np.finfo(self._delta).max,
                self.irgn_par["delta_max"])
        self.irgn_par["gamma"] = np.maximum(
            self._gamma * self.irgn_par["gamma_dec"]**ign,
            self.irgn_par["gamma_min"])/self.irgn_par["lambd"]
        self.irgn_par["omega"] = np.maximum(
            self._omega * self.irgn_par["omega_dec"]**ign,
            self.irgn_par["omega_min"])/self.irgn_par["lambd"]

    def _balanceModelGradients(self, result, ign):
        scale = np.reshape(
            self._modelgrad,
            (self.par["unknowns"],
             self.par["NScan"] * self.par["NSlice"] *
             self.par["dimY"] * self.par["dimX"]))
        scale = np.linalg.norm(scale, axis=-1)
        print("Initial Norm: ", np.linalg.norm(scale))
        print("Initial Ratio: ", scale)
        if ign == 0:
            scale /= 100/np.sqrt(self.par["unknowns"])
        else:
            scale /= np.linalg.norm(scale)/np.sqrt(self.par["unknowns"])
        scale = 1 / scale
        scale[~np.isfinite(scale)] = 1
        for uk in range(self.par["unknowns"]):
            self._model.constraints[uk].update(scale[uk])
            result[uk, ...] *= self._model.uk_scale[uk]
            self._modelgrad[uk] /= self._model.uk_scale[uk]
            self._model.uk_scale[uk] *= scale[uk]
            result[uk, ...] /= self._model.uk_scale[uk]
            self._modelgrad[uk] *= self._model.uk_scale[uk]
        scale = np.reshape(
            self._modelgrad,
            (self.par["unknowns"],
             self.par["NScan"] * self.par["NSlice"] *
             self.par["dimY"] * self.par["dimX"]))
        scale = np.linalg.norm(scale, axis=-1)
        print("Norm after rescale: ", np.linalg.norm(scale))
        print("Ratio after rescale: ", scale)

###############################################################################
# New .hdf5 save files ########################################################
###############################################################################
    def _saveToFile(self, myit, result):
        f = h5py.File(self.par["outdir"]+"output_" + self.par["fname"] + ".h5",
                      "a")
        if self._reg_type == 'TGV':
            f.create_dataset("tgv_result_iter_"+str(myit), result.shape,
                             dtype=self._DTYPE, data=result)
            f.attrs['res_tgv_iter_'+str(myit)] = self._fval
        else:
            f.create_dataset("tv_result_"+str(myit), result.shape,
                             dtype=self._DTYPE, data=result)
            f.attrs['res_tv_iter_'+str(myit)] = self._fval
        f.close()

###############################################################################
# Precompute constant terms of the GN linearization step ######################
# input: linearization point x ################################################
# iters: number of innner iterations iters ####################################
# data: the input data ########################################################
# TV: bool to switch between TV (1) and TGV (0) regularization ################
# output: optimal value of x for the inner GN step ############################
###############################################################################
###############################################################################
    def _irgnSolve3D(self, x, iters, data, GN_it):
        b = self._calcResidual(x, data, GN_it)
        if self._streamed:
            x = np.require(np.swapaxes(x, 0, 1), requirements='C')
            res = data - b + self._MRI_operator.fwdoop(
                [[x, self._coils, self._modelgrad]])
        else:
            tmpx = clarray.to_device(self._queue[0], x)
            res = data - b + self._MRI_operator.fwdoop(
                [tmpx, self._coils, self._modelgrad]).get()
            del tmpx

        tmpres = self._pdop.run((x, self._v), res, iters)
        for key in tmpres:
            if key == 'x':
                if isinstance(tmpres[key], np.ndarray):
                    x = tmpres["x"]
                else:
                    x = tmpres["x"].get()
            if key == 'v':
                if isinstance(tmpres[key], np.ndarray):
                    self._v = tmpres["v"]
                else:
                    self._v = tmpres["v"].get()
        if self._streamed:
            x = np.require(np.swapaxes(x, 0, 1), requirements='C')

        return x

    def _calcResidual(self, x, data, GN_it):
        if self._streamed:
            b, grad, sym_grad = self._calcFwdGNPartStreamed(x)
            grad_tv = grad[:, :self.par["unknowns_TGV"]]
            grad_H1 = grad[:, self.par["unknowns_TGV"]:]
        else:
            b, grad, sym_grad = self._calcFwdGNPartLinear(x)
            grad_tv = grad[:self.par["unknowns_TGV"]]
            grad_H1 = grad[self.par["unknowns_TGV"]:]
        del grad

        datacost = 1 / 2 * np.linalg.norm(data - b)**2
        # L2Cost = np.linalg.norm(x)/(2.0*self.irgn_par["delta"])
        if self._reg_type == 'TV':
            regcost = self.irgn_par["gamma"] * \
                np.sum(np.abs(grad_tv))
        elif self._reg_type == 'TGV':
            regcost = self.irgn_par["gamma"] * np.sum(
                  np.abs(grad_tv -
                         self._v)) + self.irgn_par["gamma"] * 2 * np.sum(
                             np.abs(sym_grad))
            del sym_grad
        else:
            regcost = self.irgn_par["gamma"]/2 * \
                np.linalg.norm(grad_tv)**2

        self._fval = (datacost +
                      regcost +
                       # L2Cost +
                      self.irgn_par["omega"] / 2 *
                      np.linalg.norm(grad_H1.flatten())**2)
        del grad_tv, grad_H1

        if GN_it == 0:
            self._fval_init = self._fval
            self._pdop.setFvalInit(self._fval)

        print("-" * 75)
        # print("Initial Cost: %f" % (self._fval_init))
        print("Costs of Data: %f" % (1e3*datacost / self._fval_init))
        print("Costs of T(G)V: %f" % (1e3*regcost / self._fval_init))
        # print("Costs of L2 Term: %f" % (1e3*L2Cost / self._fval_init))
        print("-" * 75)
        print("Function value at GN-Step %i: %f" %
              (GN_it, 1e3*self._fval / self._fval_init))
        print("-" * 75)
        return b

    def _calcFwdGNPartLinear(self, x):
        if self._imagespace is False:
            b = clarray.zeros(self._queue[0],
                              self._data_shape,
                              dtype=self._DTYPE)
            self._FT.FFT(b, clarray.to_device(
                self._queue[0],
                (self._step_val[:, None, ...] *
                 self.par["C"]))).wait()
            b = b.get()
        else:
            b = self._step_val

        x = clarray.to_device(self._queue[0], np.require(x, requirements="C"))
        grad = clarray.to_device(self._queue[0],
                                 np.zeros(x.shape+(4,), dtype=self._DTYPE))
        self._grad_op.fwd(
            grad,
            x,
            wait_for=grad.events +
            x.events).wait()
        x = x.get()
        grad = grad.get()
        sym_grad = None
        if self._reg_type == 'TGV':
            v = clarray.to_device(self._queue[0], self._v)
            sym_grad = clarray.to_device(self._queue[0],
                                         np.zeros(x.shape+(8,),
                                                  dtype=self._DTYPE))

            self._symgrad_op.fwd(
                sym_grad,
                v,
                wait_for=sym_grad.events +
                v.events).wait()
            sym_grad = sym_grad.get()

        return b, grad, sym_grad

    def _calcFwdGNPartStreamed(self, x):
        x = np.require(np.swapaxes(x, 0, 1), requirements='C')
        if self._imagespace is False:
            b = np.zeros(self._data_shape_T, dtype=self._DTYPE)
            if self._SMS is True:
                self._MRI_operator.FTstr.eval(
                    [b],
                    [[np.expand_dims(self._step_val, self._expdim_dat) *
                      np.expand_dims(self.par["C"], self._expdim_C)]])
                b = np.require(
                    np.transpose(
                        b,
                        self._data_trans_axes),
                    requirements='C')
            else:
                self._MRI_operator.FTstr.eval(
                    [b],
                    [[np.expand_dims(self._step_val, self._expdim_dat) *
                      np.expand_dims(self._coils, self._expdim_C)]])
        else:
            b = self._step_val
        grad = np.zeros(x.shape+(4,), dtype=self._DTYPE)
        self._grad_op.fwd([grad], [[x]])

        sym_grad = None
        if self._reg_type == 'TGV':
            sym_grad = np.zeros(x.shape+(8,), dtype=self._DTYPE)
            self._symgrad_op.fwd([sym_grad], [[self._v]])

        return b, grad, sym_grad



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module holding the simple image model for image reconstruction."""
import numpy as np
from pyqmri.models.template import BaseModel, constraints


class Model(BaseModel):
    """Image reconstruction model for MRI.

    A simple linear image model to perform image reconstruction with
    joint regularization on all Scans.

    Parameters
    ----------
      par : dict
        A python dict containing the necessary information to
        setup the object. Needs to contain the sequence related parametrs,
        e.g. TR, TE, TI, to fully describe the acquisitio process
    Attributes
    ----------
      guess : numpy.array, None
        Initial guess for the images. Set after object creation using
        "computeInitialGuess"
    """

    def __init__(self, par):
        super().__init__(par)

        par["unknowns_TGV"] = self.NScan
        par["unknowns_H1"] = 0
        par["unknowns"] = par["unknowns_TGV"]+par["unknowns_H1"]

        for j in range(par["unknowns"]):
            self.uk_scale.append(1)

        for j in range(par["unknowns"]):
            self.constraints.append(
                constraints(-100 / self.uk_scale[j],
                            100 / self.uk_scale[j],
                            False))
            
        self.guess = None

    def rescale(self, x):
        """Rescale the unknowns with the scaling factors.

        Rescales each unknown with the corresponding scaling.

        Parameters
        ----------
          x : numpy.array
            The array of unknowns to be rescaled

        Returns
        -------
          numpy.array:
            The rescaled unknowns
        """
        tmp_x = np.copy(x)
        uk_names = []
        for j in range(self.NScan):
            tmp_x[j] *= self.uk_scale[j]
            uk_names.append("Image_"+str(j))

        const = []
        for constrained in self.constraints:
            const.append(constrained.real)

        return {"data": tmp_x,
                "unknown_name": uk_names,
                "real_valued": const}

    def _execute_forward_3D(self, x):
        S = np.zeros_like(x)
        for j in range(self.NScan):
            S[j, ...] = x[j, ...] * self.uk_scale[j]
        S[~np.isfinite(S)] = 1e-20
        return S

    def _execute_gradient_3D(self, x):
        grad_M0 = np.zeros(((self.NScan, )+x.shape), dtype=self._DTYPE)
        for j in range(self.NScan):
            grad_M0[j, j, ...] = self.uk_scale[j]*np.ones_like(x[j])
        grad_M0[~np.isfinite(grad_M0)] = 1e-20
        return grad_M0

    def computeInitialGuess(self, *args):
        """Initialize unknown array for the fitting.

        This function provides an initial guess for the fitting.

        Parameters
        ----------
          args : list of objects
            Assumes the images series at position 0 and uses it as initial
            guess.
        """
        self.guess = np.zeros_like((args[0]).astype(self._DTYPE))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Density compensation for non-uniform acquried data."""
import numpy as np


def cmp(k):
    """Golden angle density compensation.

    Simple linear ramp based density compensation function

    Parameters
    ----------
      k : numpy.array
        Trajectory which should be density compensated.

    Returns
    -------
      numpy.array :
          The density compensation array.
    """
    if len(np.shape(k)) == 3:
        nspokes, N, _ = np.shape(k)
    elif len(np.shape(k)) == 4:
        _, nspokes, N, _ = np.shape(k)
    else:
        return -5

    w = np.abs(np.linspace(-N / 2, N / 2, N))  # -N/2 N/2
    w = w * (np.pi / 4) / nspokes  # no scaling seems to work better??
    w = np.repeat(w, nspokes, 0)
    w = np.reshape(w, (N, nspokes)).T

    return (w)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from pyqmri.models.template import BaseModel, constraints
plt.ion()



class Model(BaseModel):
    def __init__(self, par):
        super().__init__(par)
        self.constraints = []
        self.t = par["t"]
        self.b = par["b"]
        if "b0" in par.keys():
            self.b0 = par["b0"]
        else:
            self.b0 = self.b[0]

        if len(self.t.shape) < 2:
            self.b = self.b[None]
            self.t = self.t[None]

        self.numT1Scale = len(self.b)
        self.numC = 1#len(self.b)
        self.numAlpha = len(self.b)

        par["unknowns_TGV"] = self.numC + self.numAlpha + len(self.b)
        par["unknowns_H1"] = 0
        par["unknowns"] = par["unknowns_TGV"]+par["unknowns_H1"]


        self.unknowns = par["unknowns"]
        
        t1min = 10 #np.min(self.t)/3

        for j in range(par["unknowns"]):
            self.uk_scale.append(1)
            
        for j in range(self.numC):
            self.constraints.append(
                constraints(1e-5,
                            1e10,
                            False))
        for j in range(self.numAlpha):
            self.constraints.append(
                constraints(0.01,
                            1.1,
                            False))
        for j in range(self.numT1Scale):
            self.constraints.append(
                constraints(t1min,
                            2000/(j+1),
                            True))
        self._ind1 = 0
        self._ind2 = 0
        self._labels = []
        for j in range(len(self.b)):
            self._labels.append(
                "Field "+str(np.round(self.b[j]*1e3, 2))+" mT")
        par["weights"] = np.array([1]*self.numC+self.numAlpha*[10]+self.numT1Scale*[1],dtype=par["DTYPE_real"])
        # par["weights"] /= np.sum(par["weights"])

    def rescale(self, x):
        tmp_x = np.copy(x)
        ukname = []
        for j in range(self.numC):
            tmp_x[j] *= self.uk_scale[j]
            ukname.append("C_"+str(1+j))
        for j in range(self.numAlpha):
            tmp_x[self.numC+j] *= self.uk_scale[self.numC+j]
            ukname.append("alpha_"+str(1+j))
        for j in range(self.numT1Scale):
            tmp_x[-self.numT1Scale+j] *= self.uk_scale[-self.numT1Scale+j]
            ukname.append("T1_"+str(1+j))
        const = []
        for constrained in self.constraints:
            const.append(constrained.real)
        return {"data": tmp_x,
                "unknown_name": ukname,
                "real_valued": const}


    def _execute_forward_3D(self, x):
        S = np.zeros(
            (self.NScan, self.NSlice, self.dimY, self.dimX),
            dtype=self._DTYPE)
        t = self.t[0][:, None, None, None]


        for j in range(len(self.b)):
            offset = len(self.t[j])
            t = self.t[j][:, None, None, None]
            S[offset*(j):offset*(j+1)] = (
                x[np.mod(j,self.numC)] * self.uk_scale[np.mod(j,self.numC)]
                * (-self.b0 * x[self.numC+np.mod(j,self.numAlpha)]*self.uk_scale[self.numC+np.mod(j,self.numAlpha)] *
                   np.exp(-t / (x[-self.numT1Scale+j] * self.uk_scale[-self.numT1Scale+j]))
                   + (1 - np.exp(-t / (x[-self.numT1Scale+j] * self.uk_scale[-self.numT1Scale+j])))
                   * self.b[j])
                )
        S[~np.isfinite(S)] = 0
        S = np.array(S, dtype=self._DTYPE)
        return S

    def _execute_gradient_3D(self, x):

        gradC = self._gradC(x)
        gradAlpha = self._gradAlpha(x)
        gradT1 = self._gradT1(x)

        grad = np.concatenate((gradC, gradAlpha, gradT1), axis=0)
        return grad

    def _gradC(self, x):
        grad = np.zeros(
            (self.numC, self.NScan, self.NSlice, self.dimY, self.dimX),
            dtype=self._DTYPE)
        t = self.t[0][:, None, None, None]


        for j in range(len(self.b)):
            offset = len(self.t[j])
            t = self.t[j][:, None, None, None]
            grad[np.mod(j,self.numC), offset*(j):offset*(j+1)] = (
                self.uk_scale[np.mod(j,self.numC)]
                * (-self.b0 * x[self.numC+np.mod(j,self.numAlpha)]*self.uk_scale[self.numC+np.mod(j,self.numAlpha)] *
                   np.exp(-t / (x[-self.numT1Scale+j] * self.uk_scale[-self.numT1Scale+j]))
                   + (1 - np.exp(-t / (x[-self.numT1Scale+j] * self.uk_scale[-self.numT1Scale+j])))
                   * self.b[j])
                )
        grad[~np.isfinite(grad)] = 0

        return grad

    def _gradAlpha(self, x):
        grad = np.zeros(
            (self.numAlpha, self.NScan, self.NSlice, self.dimY, self.dimX),
            dtype=self._DTYPE)
        t = self.t[0][:, None, None, None]


        for j in range(len(self.b)):
            offset = len(self.t[j])
            t = self.t[j][:, None, None, None]
            grad[np.mod(j,self.numAlpha), offset*(j):offset*(j+1)] = (
                x[np.mod(j,self.numC)] * self.uk_scale[np.mod(j,self.numC)]*self.uk_scale[self.numC+np.mod(j,self.numAlpha)]
                * (-self.b0 *
                   np.exp(-t / (x[-self.numT1Scale+j] * self.uk_scale[-self.numT1Scale+j]))
                   )
                )
        grad[~np.isfinite(grad)] = 0

        return grad

    def _gradT1(self, x):
        grad = np.zeros(
            (self.numT1Scale, self.NScan, self.NSlice, self.dimY, self.dimX),
            dtype=self._DTYPE)
        t = self.t[0][:, None, None, None]
        for j in range(len(self.b)):
            offset = len(self.t[j])
            t = self.t[j][:, None, None, None]
            grad[j, (j)*offset:(j+1)*offset] = (
                x[np.mod(j,self.numC)]*self.uk_scale[np.mod(j,self.numC)]*(
                    -self.b0*t*x[self.numC+np.mod(j,self.numAlpha)]*self.uk_scale[self.numC+np.mod(j,self.numAlpha)]
                    * np.exp(-t/(x[-self.numT1Scale+j]*self.uk_scale[-self.numT1Scale+j]))
                    - self.b[j]*t
                    * np.exp(-t/(x[-self.numT1Scale+j]*self.uk_scale[-self.numT1Scale+j]))
                    )
                )/(x[-self.numT1Scale+j]**2*self.uk_scale[-self.numT1Scale+j])
        grad[~np.isfinite(grad)] = 0

        return grad

    def plot_unknowns(self, x, dim_2D=False):
        unknowns = self.rescale(x)
        tmp_x = unknowns["data"]
        uknames = unknowns["unknown_name"]

        images = np.abs(self._execute_forward_3D(x) / self.dscale)
        images = np.reshape(images, self.t.shape+images.shape[-3:])

        tmp_x[:self.numC] = np.abs(tmp_x[:self.numC])/self.dscale
        tmp_x[self.numC:self.numC+self.numAlpha] = np.abs(tmp_x[self.numC:self.numC+self.numAlpha])
        tmp_x[-self.numT1Scale:] = np.abs(tmp_x[-self.numT1Scale:])
        tmp_x = np.real(tmp_x)

        if dim_2D:
            pass
        else:
            if not self._figure:
                self.ax = []
                plot_dim = int(np.ceil(np.sqrt(len(self.uk_scale))))
                plt.ion()
                self._figure = plt.figure(figsize=(12, 6))
                self._figure.subplots_adjust(hspace=0.3, wspace=0)
                wd_ratio = np.tile([1, 1 / 20, 1 / (5)], plot_dim)
                self.gs = gridspec.GridSpec(
                    plot_dim+1, 3*plot_dim,
                    width_ratios=wd_ratio, hspace=0.3, wspace=0)
                self._figure.tight_layout()
                self._figure.patch.set_facecolor(plt.cm.viridis.colors[0])
                for grid in self.gs:
                    self.ax.append(plt.subplot(grid))
                    self.ax[-1].axis('off')
                self._plot = []
                for j in range(len(self.uk_scale)):
                    self._plot.append(
                        self.ax[3 * j].imshow(
                            tmp_x[j, int(self.NSlice / 2), ...]))
                    self.ax[3 *
                            j].set_title(uknames[j], color='white')
                    self.ax[3 * j + 1].axis('on')
                    cbar = self._figure.colorbar(
                        self._plot[j], cax=self.ax[3 * j + 1])
                    cbar.ax.tick_params(labelsize=12, colors='white')
                    for spine in cbar.ax.spines:
                        cbar.ax.spines[spine].set_color('white')
                plt.draw()
                plt.pause(1e-10)
                self._figure.canvas.mpl_connect(
                    'button_press_event',
                    self.onclick)

                self.plot_ax = plt.subplot(self.gs[-1, :])
                self.plot_ax.set_title("Time course", color='w')
                self.time_course_ref = []
                for j in range(len(self.b)):
                    self.time_course_ref.append(self.plot_ax.plot(
                        self.t[j], np.real(
                            self.images[j, :,
                                        int(self.NSlice/2),
                                        self._ind2, self._ind1]).T,
                        'x', label=self._labels[j])[0])
                self.plot_ax.set_prop_cycle(None)
                legend = self.plot_ax.legend(frameon=True, framealpha=0.3)
                for _txt in legend.texts:
                    _txt.set_alpha(0.3)
                for lh in legend.legendHandles:
                    lh._legmarker.set_alpha(0.3)
                self.time_course = self.plot_ax.plot(
                    self.t.T, np.real(
                        images[..., int(self.NSlice/2),
                               self._ind2, self._ind1]).T)
                self.plot_ax.set_ylim(
                    np.minimum(np.real(images[...,
                                              int(self.NSlice/2),
                                              self._ind2,
                                              self._ind1]).min(),
                               np.real(self.images[...,
                                                   int(self.NSlice/2),
                                                   self._ind2,
                                                   self._ind1]).min()),
                    1.2*np.maximum(np.real(images[...,
                                                  int(self.NSlice/2),
                                                  self._ind2,
                                                  self._ind1]).max(),
                                   np.real(self.images[...,
                                                       int(self.NSlice/2),
                                                       self._ind2,
                                                       self._ind1]).max()))
                for spine in self.plot_ax.spines:
                    self.plot_ax.spines[spine].set_color('white')
                self.plot_ax.xaxis.label.set_color('white')
                self.plot_ax.yaxis.label.set_color('white')
                self.plot_ax.tick_params(axis='both', colors='white')

                plt.draw()
                plt.show()
                plt.pause(1e-4)
            else:
                for j in range(len(self.uk_scale)):
                    self._plot[j].set_data(
                        tmp_x[j, int(self.NSlice / 2), ...])
                    self._plot[j].set_clim(
                        [tmp_x[j].min(), tmp_x[j].max()])

                for j in range(len(self.b)):
                    self.time_course[j].set_ydata(
                        np.real(images[
                            j, :, int(self.NSlice/2), self._ind2, self._ind1]))
                self.plot_ax.set_ylim(
                    np.minimum(np.real(images[...,
                                              int(self.NSlice/2),
                                              self._ind2,
                                              self._ind1]).min(),
                               np.real(self.images[...,
                                                   int(self.NSlice/2),
                                                   self._ind2,
                                                   self._ind1]).min()),
                    1.2*np.maximum(np.real(images[...,
                                                  int(self.NSlice/2),
                                                  self._ind2,
                                                  self._ind1]).max(),
                                   np.real(self.images[...,
                                                       int(self.NSlice/2),
                                                       self._ind2,
                                                       self._ind1]).max()))
                plt.draw()
                plt.pause(1e-10)

    def onclick(self, event):
        if event.inaxes in self.ax[::3]:
            self._ind1 = int(event.xdata)
            self._ind2 = int(event.ydata)
            for j in range(len(self.b)):
                self.time_course_ref[j].set_ydata(np.real(
                        self.images[j, :,
                                    int(self.NSlice/2),
                                    self._ind2, self._ind1]).T)

    def computeInitialGuess(self, *args):
        self.dscale = args[1]
        self.images = np.reshape(np.abs(args[0]/args[1]),
                                 self.t.shape+args[0].shape[-3:])
        test_M0 = []
        for j in range(self.numC):
            test_M0.append(0.1*np.ones(
                (self.NSlice, self.dimY, self.dimX), dtype=self._DTYPE)*
                np.exp(1j*np.angle(args[0][0])))
            self.constraints[j].update(1/args[1])
        test_Xi = []
        for j in range(self.numAlpha):
            test_Xi.append(
                1 *
                np.ones(
                    (self.NSlice, self.dimY, self.dimX), dtype=self._DTYPE))
 
        test_R1 = []
        # self.b *= args[1]
        for j in range(self.numT1Scale):
            test_R1.append(
                50 *
                np.ones(
                    (self.NSlice, self.dimY, self.dimX), dtype=self._DTYPE))
        x = np.array(

            test_M0 + test_Xi + test_R1, dtype=self._DTYPE)
        x_scale = np.max(np.abs(x).reshape(x.shape[0], -1), axis=-1)
        self.uk_scale = x_scale
        self.guess = x/x_scale[:,None,None,None]
        for uk in range(self.unknowns):
            self.constraints[uk].update(x_scale[uk])

        # self.guess = np.array(
        #     test_M0 + test_Xi + test_R1, dtype=self._DTYPE)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module holding the variable flip angle model for T1 fitting."""
import numpy as np
from pyqmri.models.template import BaseModel, constraints


class Model(BaseModel):
    """B1 model for MRI parameter quantification.

    This class holds a variable flip angle model for T1 quantification from
    complex MRI data. It realizes a forward application of the analytical
    signal expression and the partial derivatives with respesct to
    each parameter of interest, as required by the abstract methods in
    the BaseModel. The fitting target is the exponential term itself
    which is easier to fit than the corresponding timing constant.

    The rescale function applies a transformation and returns the expected
    T1 values in ms.

    Parameters
    ----------
      par : dict
        A python dict containing the necessary information to
        setup the object. Needs to contain the sequence related parametrs,
        e.g. TR, TE, TI, to fully describe the acquisitio process

    Attributes
    ----------
      TR : float
        Repetition time of the gradient echo sequence.
      fa : numpy.array
        A vector containing all flip angles, one per scan.
      uk_scale : list of float
        Scaling factors for each unknown to balance the partial derivatives.
      guess : numpy.array, None
        The initial guess. Needs to be set using "computeInitialGuess"
        prior to fitting.
    """

    def __init__(self, par):
        super().__init__(par)


        par["unknowns_TGV"] = 4
        par["unknowns_H1"] = 0
        par["unknowns"] = par["unknowns_TGV"]+par["unknowns_H1"]


        for j in range(par["unknowns"]):
            self.uk_scale.append(1)

        self.constraints.append(
            constraints(-1e5,
                        1e5,
                        True))
        self.constraints.append(
            constraints(-1e5,
                        1e5,
                        True))
        self.constraints.append(
            constraints(-1e5,
                        1e5,
                        True))
        self.constraints.append(
            constraints(-1e5,
                        1e5,
                        True))

    def rescale(self, x):
        """Rescale the unknowns with the scaling factors.

        Rescales each unknown with the corresponding scaling factor and
        applies a transformation for the time constants of the exponentials,
        yielding a resulting T1 in milliseconds.

        Parameters
        ----------
          x : numpy.array
            The array of unknowns to be rescaled

        Returns
        -------
          numpy.array:
            The rescaled unknowns
        """
        tmp_x = np.copy(x)
        for j in range(x.shape[0]):
            tmp_x[j] *= self.uk_scale[j]
        const = []
        for constrained in self.constraints:
            const.append(constrained.real)
        tmp2 = np.copy(x)
        tmp2[0] = np.sqrt(tmp_x[0]**2+tmp_x[1]**2)
        tmp2[1] = np.angle(tmp_x[0]+1j*tmp_x[1])
        tmp2[2] = np.sqrt(tmp_x[2]**2+tmp_x[3]**2)
        tmp2[3] = np.angle(tmp_x[2]+1j*tmp_x[3])
        tmp2[3] = -(tmp2[1] - tmp2[3])
        return {"data": tmp2,
                "unknown_name": ["R", "P", "R", "P"],
                "real_valued": const}
    

    def _execute_forward_3D(self, x):
        
        S1 = x[0]*self.uk_scale[0] + 1j *(x[1]*self.uk_scale[1])
        S2 = x[2]*self.uk_scale[2] + 1j *(x[3]*self.uk_scale[3])
        S3 = (x[0]*self.uk_scale[0]-x[2]*self.uk_scale[2]) + 1j*(x[1]*self.uk_scale[1]-x[3]*self.uk_scale[3])
        
        S = np.array([S1,S2,S3], dtype=self._DTYPE)
        S[~np.isfinite(S)] = 1e-20
        S = np.array(S, dtype=self._DTYPE)
        return S

    def _execute_gradient_3D(self, x):
        
        grad_r1_p1 = self.uk_scale[0] * np.ones_like(x[0])
        grad_r1_p2 = np.zeros_like(grad_r1_p1)
        grad_r1_p3 = self.uk_scale[0] * np.ones_like(x[0])
        
        grad_r1 = np.array([grad_r1_p1, grad_r1_p2, grad_r1_p3])

        grad_r2_p1 = 1j*self.uk_scale[1] * np.ones_like(x[0])
        grad_r2_p2 = np.zeros_like(grad_r1_p1)
        grad_r2_p3 = 1j*self.uk_scale[1] * np.ones_like(x[0])
        
        grad_r2 = np.array([grad_r2_p1, grad_r2_p2, grad_r2_p3])
        
        
        grad_r3_p1 = self.uk_scale[2] * np.ones_like(x[2])
        grad_r3_p2 = np.zeros_like(grad_r1_p1)
        grad_r3_p3 = -self.uk_scale[2] * np.ones_like(x[2])
        
        grad_r3 = np.array([grad_r3_p1, grad_r3_p2, grad_r3_p3])

        grad_r4_p1 = 1j*self.uk_scale[3] * np.ones_like(x[3])
        grad_r4_p2 = np.zeros_like(grad_r1_p1)
        grad_r4_p3 = -1j*self.uk_scale[3] * np.ones_like(x[3])
        
        grad_r4 = np.array([grad_r4_p1, grad_r4_p2, grad_r4_p3])
                
        grad = np.array([grad_r1, grad_r2, grad_r3, grad_r4], dtype=self._DTYPE)
        grad[~np.isfinite(grad)] = 1e-20
        return grad

    def computeInitialGuess(self, *args):
        """Initialize unknown array for the fitting.

        This function provides an initial guess for the fitting.

        Parameters
        ----------
          args : list of objects
            Serves as universal interface. No objects need to be passed
            here.
        """
        test_T1 = 1 * np.ones(
            (self.NSlice, self.dimY, self.dimX), dtype=self._DTYPE)
        test_M0 = 1 * np.ones(
            (self.NSlice, self.dimY, self.dimX),
            dtype=self._DTYPE)

        x = np.array([np.real(args[0][0]),
                      np.imag(args[0][0]), np.real(args[0][1]), np.imag(args[0][1])], dtype=self._DTYPE)
        self.guess = x



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from pyqmri.models.template import BaseModel, constraints
import numexpr as ne
plt.ion()


def _expAttT1r(del_t, T1r):
    return ne.evaluate(
        "exp(del_t*T1r)")

def _expT1app(t, T1app):
    return ne.evaluate(
        "exp(-t*T1app)")

def _expT1r(t, T1r):
    return ne.evaluate(
        "exp(t*T1r)")


def _T1pr(T1, f, f_sc, lambd):
    return ne.evaluate(
        "1/T1+f*f_sc/lambd")

def _T1r(T1app, T1b):
    return ne.evaluate(
        "T1app - 1/T1b")


def _S1(M0, alpha, lambd, f, T1r, expT1r, expAttT1r, expT1app):
    return ne.evaluate(
        "2*M0*alpha*f*(expT1r - expAttT1r)*expT1app/(lambd*T1r)")


def _S2(M0, alpha, lambd, f, T1r, del_t, tau, expAttT1r, expT1app):
    return ne.evaluate(
        "2*M0*alpha*f*(exp((del_t + tau)*T1r) - expAttT1r)*expT1app/(lambd*T1r)")


def _delCBF1(M0, alpha, lambd, f, f_sc, del_t,
             del_t_sc, t, T1r, expT1r, expT1app, expAttT1r):
    return ne.evaluate(
        "-2*M0*alpha*f*f_sc**2*t*(expT1r - expAttT1r)*expT1app/(lambd**2*T1r) \
         - 2*M0*alpha*f*f_sc**2*(expT1r - expAttT1r)*expT1app/(lambd**2*T1r**2) \
         + 2*M0*alpha*f*f_sc*(-del_t*del_t_sc*f_sc*expAttT1r/lambd + f_sc*t*expT1r/lambd)*expT1app/(lambd*T1r) \
         + 2*M0*alpha*f_sc*(expT1r - expAttT1r)*expT1app/(lambd*T1r)")


def _delCBF2(M0, alpha, lambd, f, f_sc, del_t,
             del_t_sc, t, tau, expT1app, expAttT1r, T1r):
    return ne.evaluate(
        "-2*M0*alpha*f*f_sc**2*t*(exp((del_t*del_t_sc + tau)*T1r) - expAttT1r)*expT1app/(lambd**2*T1r) \
         - 2*M0*alpha*f*f_sc**2*(exp((del_t*del_t_sc + tau)*T1r) - expAttT1r)*expT1app/(lambd**2*T1r**2) \
         + 2*M0*alpha*f*f_sc*(-del_t*del_t_sc*f_sc*expAttT1r/lambd \
                              + f_sc*(del_t*del_t_sc + tau)*exp((del_t*del_t_sc + tau)*T1r)/lambd)*expT1app/(lambd*T1r) \
         + 2*M0*alpha*f_sc*(exp((del_t*del_t_sc + tau)*T1r) - expAttT1r)*expT1app/(lambd*T1r)")


def _delATT1(M0, alpha, lambd, f, f_sc, 
             del_t_sc, expT1app, expAttT1r):
    return ne.evaluate(
        "-2*M0*alpha*del_t_sc*f*f_sc*expT1app*expAttT1r/lambd")


def _delATT2(M0, alpha, lambd, f, f_sc, del_t,
             del_t_sc, T1r, tau, expT1app, expAttT1r):
    return ne.evaluate(
        "2*M0*alpha*f*f_sc*(del_t_sc*T1r*exp((del_t*del_t_sc + tau)*T1r) - del_t_sc*T1r*expAttT1r)*expT1app/(lambd*T1r)")


class Model(BaseModel):
    def __init__(self, par):
        super().__init__(par)
        full_slices = par["file"]["T1b"].shape[0]
        sliceind = slice(int(full_slices / 2) -
                         int(np.floor((par["NSlice"]) / 2)),
                         int(full_slices / 2) +
                         int(np.ceil(par["NSlice"] / 2)))
        self.T1b = par["file"]["T1b"][sliceind]
        self.T1 = par["file"]["T1"][sliceind]
        self.lambd = par["file"]["lambd"][sliceind]
        self.M0 = par["file"]["M0"][sliceind]
        self.tau = par["file"]["tau"][:, sliceind]
        self.t = par['t']
        
        while len(self.t.shape) < 4:
            self.t = self.t[..., None]
        self.t = np.ones(self.M0.shape)*self.t
        self.alpha = par["file"]["alpha"][sliceind]

        par["unknowns_TGV"] = 2
        par["unknowns_H1"] = 0
        par["unknowns"] = par["unknowns_TGV"]+par["unknowns_H1"]
        self.unknowns = par["unknowns"]

        for j in range(self.unknowns):
            self.uk_scale.append(1)

        self.constraints.append(
            constraints(0,
                        200,
                        False))
        self.constraints.append(
            constraints(0.01/60,
                        5/60,
                        True))
        self._ind1 = 35
        self._ind2 = 46

    def _execute_forward_2D(self, x, islice):
        print("2D Functions not implemented")
        raise NotImplementedError

    def _execute_gradient_2D(self, x, islice):
        print("2D Functions not implemented")
        raise NotImplementedError

    def _execute_forward_3D(self, x):
        f = x[0, ...] * self.uk_scale[0]
        del_t = x[1, ...] * self.uk_scale[1]

        S = np.zeros((self.NScan, self.NSlice, self.dimY, self.dimX),
                     dtype=self._DTYPE)

        T1appinv = _T1pr(self.T1, x[0], self.uk_scale[0], self.lambd)
        T1rinv = _T1r(T1appinv, self.T1b)
        
        expT1r = _expT1r(self.t, T1rinv)
        expT1app = _expT1app(self.t, T1appinv)
        expAttT1r = _expAttT1r(del_t, T1rinv)
        
        for j in range(self.t.shape[0]):
            ind_low = self.t[j] >= del_t
            ind_high = self.t[j] < (del_t+self.tau[j])
            ind = ind_low & ind_high
            if np.any(ind):
                S[j, ind] = _S1(self.M0[ind], self.alpha[ind],
                                self.lambd[ind], f[ind], 
                                T1rinv[ind],
                                expT1r[j, ind], expAttT1r[ind],
                                expT1app[j, ind])
                
            ind = self.t[j] >= del_t + self.tau[j]
            if np.any(ind):
                S[j, ind] = _S2(self.M0[ind], self.alpha[ind],
                                self.lambd[ind], f[ind], T1rinv[ind],
                                del_t[ind], 
                                self.tau[j, ind], 
                                expAttT1r[ind],
                                expT1app[j, ind])
        S[~np.isfinite(S)] = 1e-20
        S = np.array(S, dtype=self._DTYPE)
        return S

    def _execute_gradient_3D(self, x):
        f_sc = self.uk_scale[0]
        del_t_sc = self.uk_scale[1]
        del_t = x[1]*del_t_sc
        grad = np.zeros((self.unknowns, self.NScan,
                         self.NSlice, self.dimY, self.dimX), dtype=self._DTYPE)
        t = self.t
        T1appinv = _T1pr(self.T1, x[0], self.uk_scale[0], self.lambd)
        T1rinv = _T1r(T1appinv, self.T1b)
        
        expT1r = _expT1r(t, T1rinv)
        expT1app = _expT1app(t, T1appinv)
        expAttT1r = _expAttT1r(del_t, T1rinv)
   
        
        for j in range(self.t.shape[0]):
            ind_low = self.t[j] >= del_t
            ind_high = self.t[j] < (del_t+self.tau[j])
            ind = ind_low & ind_high
            if np.any(ind):
                grad[0, j, ind] = _delCBF1(self.M0[ind], self.alpha[ind],
                                           self.lambd[ind], x[0, ind],
                                           f_sc,
                                           x[1, ind],
                                           del_t_sc, t[j, ind],
                                           T1rinv[ind],
                                           expT1r[j, ind],
                                           expT1app[j, ind],
                                           expAttT1r[ind])
                grad[1, j, ind] = _delATT1(self.M0[ind], self.alpha[ind],
                                           self.lambd[ind], x[0, ind],
                                           f_sc, 
                                           del_t_sc, 
                                           expT1app[j, ind],
                                           expAttT1r[ind])

            ind = self.t[j] >= del_t + self.tau[j]
            if np.any(ind):
                grad[0, j, ind] = _delCBF2(self.M0[ind], self.alpha[ind],
                                           self.lambd[ind], x[0, ind],
                                           f_sc, x[1, ind],
                                           del_t_sc, 
                                           t[j, ind],
                                           self.tau[j, ind],
                                           expT1app[j, ind],
                                           expAttT1r[ind],
                                           T1rinv[ind])
                grad[1, j, ind] = _delATT2(self.M0[ind], self.alpha[ind],
                                           self.lambd[ind], x[0, ind],
                                           f_sc,
                                           x[1, ind],
                                           del_t_sc, 
                                           T1rinv[ind], 
                                           self.tau[j, ind],
                                           expT1app[j, ind],
                                           expAttT1r[ind])
        grad[~np.isfinite(grad)] = 1e-20
        grad = np.array(grad, dtype=self._DTYPE)
        return grad

    def plot_unknowns(self, x, dim_2D=False):
        unknowns = self.rescale(x)
        tmp_x = unknowns["data"]
        
        images = self._execute_forward_3D(x) / self.dscale
        
        f = np.abs(tmp_x[0, ...]/ self.dscale)
        
        del_t = np.abs(tmp_x[1, ...])*60
#        del_t[f <= 15] = 0
        f_min = f.min()
        f_max = f.max()
        del_t_min = del_t.min()
        del_t_max = del_t.max()
        # ind1 = 46
        # ind2 = 35# int(images.shape[-1]/2) 30, 60
        off = 0
        [z, y, x] = f.shape
        if len(self.t.shape) == 4:
            t = self.t[:, int(self.NSlice/2), 0, 0]
        else:
            t = self.t
        if dim_2D:
            pass
        else:
            if not self._figure:
                self.ax = []
                plt.ion()
                self._figure = plt.figure(figsize=(12, 6))
                self._figure.subplots_adjust(hspace=0, wspace=0)
                self.gs = gridspec.GridSpec(
                    4, 6,
                    width_ratios=[
                        x / (20 * z), x / z, 1, x / z, 1, x / (20 * z)],
                    height_ratios=[x / z, 1, x / z, x / z])
                self._figure.tight_layout()
                self._figure.patch.set_facecolor(plt.cm.viridis.colors[0])
                for grid in self.gs:
                    self.ax.append(plt.subplot(grid))
                    self.ax[-1].axis('off')

                self.ax[1].volume = f
                self.ax[1].index = int(self.NSlice / 2)
                self.f_plot = self.ax[1].imshow(
                    (f[int(self.NSlice / 2), ...]))
                self.ax[7].volume = np.swapaxes(f, 0, 1)
                self.ax[7].index = int(f.shape[1] / 2)
                self.f_plot_cor = self.ax[7].imshow(
                    (f[:, int(f.shape[1] / 2), ...]))
                self.ax[2].volume = f.T
                self.ax[2].index = int(f.shape[-1] / 2)
                self.f_plot_sag = self.ax[2].imshow(
                    np.flip((f[:, :, int(f.shape[-1] / 2)]).T, 1))
                self.ax[1].set_title('CBF', color='white')
                self.ax[1].set_anchor('SE')
                self.ax[2].set_anchor('SW')
                self.ax[7].set_anchor('NE')
                cax = plt.subplot(self.gs[:2, 0])
                cbar = self._figure.colorbar(self.f_plot, cax=cax)
                cbar.ax.tick_params(labelsize=12, colors='white')
                cax.yaxis.set_ticks_position('left')
                for spine in cbar.ax.spines:
                    cbar.ax.spines[spine].set_color('white')

                self.ax[3].volume = del_t
                self.ax[3].index = int(self.NSlice / 2)
                self.del_t_plot = self.ax[3].imshow(
                    (del_t[int(self.NSlice / 2)+off, ...]))
                self.ax[9].volume = np.swapaxes(del_t, 0, 1)
                self.ax[9].index = int(del_t.shape[1] / 2)
                self.del_t_plot_cor = self.ax[9].imshow(
                    (del_t[:, int(del_t.shape[1] / 2), ...]))
                self.ax[4].volume = del_t.T
                self.ax[4].index = int(del_t.shape[-1] / 2)
                self.del_t_plot_sag = self.ax[4].imshow(
                    np.flip((del_t[:, :, int(del_t.shape[-1] / 2)]).T, 1))
                self.ax[3].set_title('ATT', color='white')
                self.ax[3].set_anchor('SE')
                self.ax[4].set_anchor('SW')
                self.ax[9].set_anchor('NE')
                cax = plt.subplot(self.gs[:2, 5])
                cbar = self._figure.colorbar(self.del_t_plot, cax=cax)
                cbar.ax.tick_params(labelsize=12, colors='white')
                for spine in cbar.ax.spines:
                    cbar.ax.spines[spine].set_color('white')

                plt.draw()
                plt.pause(1e-10)

                self.plot_ax = plt.subplot(self.gs[-1, :])
                self.plot_ax.set_title("Time course", color='w')
                self.time_course_ref = []

                self.time_course_ref.append(self.plot_ax.plot(
                    t*60, np.real(
                        self.images[...,
                                    int(self.NSlice/2),
                                    self._ind2, self._ind1]),
                    'x')[0])
                
                self.plot_ax.set_prop_cycle(None)

                self.time_course = self.plot_ax.plot(
                    t*60, np.real(
                        images[..., int(self.NSlice/2),
                               self._ind2, self._ind1]))
                
                self.plot_ax.set_ylim(
                    np.minimum(np.real(images[...,
                                              int(self.NSlice/2),
                                              self._ind2,
                                              self._ind1]).min(),
                               np.real(self.images[...,
                                                   int(self.NSlice/2),
                                                   self._ind2,
                                                   self._ind1]).min()),
                    1.2*np.maximum(np.real(images[...,
                                                  int(self.NSlice/2),
                                                  self._ind2,
                                                  self._ind1]).max(),
                                   np.real(self.images[...,
                                                       int(self.NSlice/2),
                                                       self._ind2,
                                                       self._ind1]).max()))
                for spine in self.plot_ax.spines:
                    self.plot_ax.spines[spine].set_color('white')
                self.plot_ax.xaxis.label.set_color('white')
                self.plot_ax.yaxis.label.set_color('white')
                self.plot_ax.tick_params(axis='both', colors='white')
                
               
                plt.draw()
                plt.pause(1e-10)
                
                self._figure.canvas.mpl_connect(
                    'button_press_event',
                    self.onclick)
                self._figure.canvas.mpl_connect(
                    'scroll_event',
                    self.onscroll)
                
            else:
                self.ax[1].volume = f
                self.ax[7].volume = np.swapaxes(f, 0, 1)
                self.ax[2].volume = f.T

                self.ax[3].volume = del_t
                self.ax[9].volume = np.swapaxes(del_t, 0, 1)
                self.ax[4].volume = del_t.T
                
                self.ax[1].images[0].set_array(self.ax[1].volume[self.ax[1].index])
                self.ax[2].images[0].set_array(self.ax[2].volume[self.ax[2].index])
                self.ax[7].images[0].set_array(self.ax[7].volume[self.ax[7].index])
                
                self.f_plot.set_clim([f_min, f_max])
                self.f_plot_cor.set_clim([f_min, f_max])
                self.f_plot_sag.set_clim([f_min, f_max])
                
                self.ax[3].images[0].set_array(self.ax[3].volume[self.ax[3].index])
                self.ax[4].images[0].set_array(self.ax[4].volume[self.ax[4].index])
                self.ax[9].images[0].set_array(self.ax[9].volume[self.ax[9].index])
                
                self.del_t_plot.set_clim([del_t_min, del_t_max])
                self.del_t_plot_sag.set_clim([del_t_min, del_t_max])
                self.del_t_plot_cor.set_clim([del_t_min, del_t_max])

                self.time_course[0].set_ydata(
                    np.real(images[:, self.ax[1].index, 
                                   self._ind2, self._ind1]))
                
                self.plot_ax.set_ylim(
                    np.minimum(np.real(images[...,
                                              self.ax[1].index,
                                              self._ind2,
                                              self._ind1]).min(),
                               np.real(self.images[...,
                                                   self.ax[1].index,
                                                   self._ind2,
                                                   self._ind1]).min()),
                    1.2*np.maximum(np.real(images[...,
                                                  self.ax[1].index,
                                                  self._ind2,
                                                  self._ind1]).max(),
                                   np.real(self.images[...,
                                                       self.ax[1].index,
                                                       self._ind2,
                                                       self._ind1]).max()))
                plt.draw()
                plt.pause(1e-10)
                
    def onclick(self, event):
        if event.inaxes in [self.ax[1], self.ax[3]]:
            self._ind1 = int(event.xdata)
            self._ind2 = int(event.ydata)

            self.time_course_ref[0].set_ydata(np.real(
                    self.images[...,
                                self.ax[1].index,
                                self._ind2, self._ind1]))
            self.plot_ax.set_ylim(
                (np.real(self.images[...,
                                     self.ax[1].index,
                                     self._ind2,
                                     self._ind1]).min()),
                1.2*(np.real(self.images[...,
                                         self.ax[1].index,
                                         self._ind2,
                                         self._ind1]).max()))
            
    def onscroll(self, event):
        if event.inaxes in [self.ax[1], self.ax[3]]:
            fig = event.canvas.figure
            ax = [self.ax[1], self.ax[3]]
        
        elif event.inaxes in [self.ax[2], self.ax[4]]:
            fig = event.canvas.figure
            ax = [self.ax[2], self.ax[4]]
                        
        elif event.inaxes in [self.ax[7], self.ax[9]]:
            fig = event.canvas.figure
            ax = [self.ax[7], self.ax[9]]
        else:
            return
        
        for i, axes in enumerate(ax):
            if axes.index is not None:
                volume = axes.volume
                if (int((axes.index - event.step) >= volume.shape[0]) or
                        int((axes.index - event.step) < 0)):
                    pass
                else:
                    ax[i].index = int((axes.index - event.step) % volume.shape[0])
                    ax[i].images[0].set_array(volume[ax[i].index])
                    fig.canvas.draw()
        plt.draw()
        plt.pause(1e-10)
                
                        
                        

    def computeInitialGuess(self, *args):
        self.dscale = args[1]
        self.constraints[0].update(1/self.dscale)
        self.images = args[0]/args[1]
        test_f = 30 * self.dscale * np.ones(
            (self.NSlice, self.dimY, self.dimX), dtype=self._DTYPE)
        test_del_t = 1/60 * np.ones(
            (self.NSlice, self.dimY, self.dimX), dtype=self._DTYPE)
        x = np.array([test_f,
                      test_del_t], dtype=self._DTYPE)
        self.guess = x
